@@ -11,7 +11,7 @@
  *
  * Configuration file goes in /etc/cachefiles.conf and is of the form:
  *
- *	dir /var/fscache
+ *	dir /var/cache/fscache
  *	tag mycache
  *	brun 10%
  *	bcull 7%
@@ -27,8 +27,11 @@
  * NUL characters are cause for error
  */
 
+#define CACHEFILESD_VERSION "0.10.7"
+
 #define _GNU_SOURCE
 #include <stdarg.h>
+#include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -43,9 +46,12 @@
 #include <dirent.h>
 #include <time.h>
 #include <poll.h>
+#include <limits.h>
+#include <grp.h>
 #include <sys/inotify.h>
 #include <sys/time.h>
 #include <sys/vfs.h>
+#include <sys/stat.h>
 
 typedef enum objtype {
 	OBJTYPE_INDEX,
@@ -62,9 +68,9 @@ struct object {
 	DIR		*dir;		/* this object's directory (or NULL for data obj) */
 	ino_t		ino;		/* inode number of this object */
 	int		usage;		/* number of users of this object */
-	char		empty;		/* T if directory empty */
-	char		new;		/* T if object new */
-	char		cullable;	/* T if object now cullable */
+	bool		empty;		/* T if directory empty */
+	bool		new;		/* T if object new */
+	bool		cullable;	/* T if object now cullable */
 	objtype_t	type;		/* type of object */
 	time_t		atime;		/* last access time on this object */
 	char		name[1];	/* name of this object */
@@ -78,11 +84,11 @@ static struct object root = {
 };
 
 static int nobjects = 1;
-static int nopendir = 0;
+static int nopendir;
 
 /* current scan point */
-static struct object *scan = &root;
-static int jumpstart_scan = 0;
+static struct object *scan_cursor;
+static bool scan_signalled, stop_signalled, reap_signalled;
 
 /* ranked order of cullable objects
  * - we have two tables: one we're building and one that's full of ready to be
@@ -92,9 +98,13 @@ static unsigned culltable_size = 4096;
 static struct object **cullbuild;
 static struct object **cullready;
 
-static int oldest_build = -1;
-static int oldest_ready = -1;
-static int ncullable = 0;
+static unsigned nr_in_build_table;
+static unsigned nr_in_ready_table;
+static int ncullable;
+static bool kernel_wants_cull;
+static bool have_nr_releases;
+static unsigned long long f_released_since_last_scan;
+static unsigned long long b_released_since_last_scan;
 
 
 static const char *configfile = "/etc/cachefilesd.conf";
@@ -103,19 +113,32 @@ static const char *procfile = "/proc/fs/cachefiles";
 static const char *pidfile = "/var/run/cachefilesd.pid";
 static char *cacheroot, *graveyardpath;
 
-static int xdebug, xnolog, xopenedlog;
-static int stop, reap, cull; //, statecheck;
+static bool culling_disabled;
+static bool xnolog, xopenedlog;
+static int xdebug;
 static int graveyardfd;
 static unsigned long long brun, bcull, bstop, frun, fcull, fstop;
+static unsigned long long b_resume_threshold = ULLONG_MAX;
+static unsigned long long f_resume_threshold = 5;
+
+static const gid_t group_list[0];
 
 #define cachefd 3
 
-static void help(void) __attribute__((noreturn));
-static void help(void)
+static __attribute__((noreturn))
+void version(void)
+{
+	printf("cachefilesd version " CACHEFILESD_VERSION "\n");
+	exit(0);
+}
+
+static __attribute__((noreturn))
+void help(void)
 {
 	fprintf(stderr,
 		"Format:\n"
 		"  /sbin/cachefilesd [-d]* [-s] [-n] [-p <pidfile>] [-f <configfile>]\n"
+		"  /sbin/cachefilesd -v\n"
 		"\n"
 		"Options:\n"
 		"  -d\tIncrease debugging level (cumulative)\n"
@@ -123,6 +146,7 @@ static void help(void)
 		"  -s\tMessage output to stderr instead of syslog\n"
 		"  -p <pidfile>\tWrite the PID into the file\n"
 		"  -f <configfile>\n"
+		"  -v\tPrint version and exit\n"
 		"\tRead the specified configuration file instead of"
 		" /etc/cachefiles.conf\n");
 
@@ -142,7 +166,7 @@ void __error(int excode, const char *fmt, ...)
 	else {
 		if (!xopenedlog) {
 			openlog("cachefilesd", LOG_PID, LOG_DAEMON);
-			xopenedlog = 1;
+			xopenedlog = true;
 		}
 
 		va_start(va, fmt);
@@ -174,7 +198,7 @@ void __message(int dlevel, int level, const char *fmt, ...)
 		else if (!xnolog) {
 			if (!xopenedlog) {
 				openlog("cachefilesd", LOG_PID, LOG_DAEMON);
-				xopenedlog = 1;
+				xopenedlog = true;
 			}
 
 			va_start(va, fmt);
@@ -197,7 +221,8 @@ static void reap_graveyard_aux(const char *dirname);
 static void read_cache_state(void);
 static int is_object_in_use(const char *filename);
 static void cull_file(const char *filename);
-static void build_cull_table(void);
+static void begin_building_cull_table(void);
+static bool build_cull_table(void);
 static void decant_cull_table(void);
 static void insert_into_cull_table(struct object *object);
 static void put_object(struct object *object);
@@ -213,9 +238,8 @@ static void cull_objects(void);
  */
 static void sigterm(int sig)
 {
-	stop = 1;
-
-} /* end sigterm() */
+	stop_signalled = true;
+}
 
 /*****************************************************************************/
 /*
@@ -223,9 +247,8 @@ static void sigterm(int sig)
  */
 static void sigio(int sig)
 {
-	reap = 1;
-
-} /* end sigio() */
+	reap_signalled = true;
+}
 
 /*****************************************************************************/
 /*
@@ -233,9 +256,8 @@ static void sigio(int sig)
  */
 static void sigalrm(int sig)
 {
-	jumpstart_scan = 1;
-
-} /* end sigalrm() */
+	scan_signalled = true;
+}
 
 /*****************************************************************************/
 /*
@@ -252,8 +274,7 @@ static void write_pidfile(void)
 	if (fprintf(pf, "%d\n", getpid()) < 0 ||
 	    fclose(pf) == EOF)
 		oserror("Unable to write PID file: %s", pidfile);
-
-} /* end write_pidfile() */
+}
 
 /*****************************************************************************/
 /*
@@ -268,14 +289,18 @@ int main(int argc, char *argv[])
 	FILE *config;
 	char *line, *cp;
 	long page_size;
-	int _cachefd, nullfd, opt, loop, open_max, nodaemon = 0;
+	int _cachefd, nullfd, opt, loop, open_max;
+	bool nodaemon = false;
 
 	/* handle help request */
 	if (argc == 2 && strcmp(argv[1], "--help") == 0)
 		help();
 
+	if (argc == 2 && strcmp(argv[1], "--version") == 0)
+		version();
+
 	/* parse the arguments */
-	while (opt = getopt(argc, argv, "dsnf:p:"),
+	while (opt = getopt(argc, argv, "dsnNf:p:v"),
 	       opt != EOF
 	       ) {
 		switch (opt) {
@@ -286,12 +311,17 @@ int main(int argc, char *argv[])
 
 		case 's':
 			/* disable syslog writing */
-			xnolog = 1;
+			xnolog = true;
 			break;
 
 		case 'n':
 			/* don't daemonise */
-			nodaemon = 1;
+			nodaemon = true;
+			break;
+
+		case 'N':
+			/* disable culling */
+			culling_disabled = true;
 			break;
 
 		case 'f':
@@ -304,6 +334,10 @@ int main(int argc, char *argv[])
 			pidfile = optarg;
 			break;
 
+		case 'v':
+			/* print the version and exit */
+			version();
+
 		default:
 			opterror("Unknown commandline option '%c'", optopt);
 		}
@@ -319,6 +353,9 @@ int main(int argc, char *argv[])
 		oserror("Unable to get max open files");
 
 	/* become owned by root */
+	if (setgroups(sizeof(group_list) / sizeof(gid_t), group_list) < 0)
+		oserror("Unable to clear the supplementary groups");
+
 	if (setresuid(0, 0, 0) < 0)
 		oserror("Unable to set UID to 0");
 
@@ -392,6 +429,12 @@ int main(int argc, char *argv[])
 		if (*cp == '#')
 			continue;
 
+		/*  allow culling to be disabled */
+		if (memcmp(cp, "nocull", 6) == 0 &&
+		    (!cp[6] || isspace(cp[6]))) {
+			culling_disabled = true;
+		}
+
 		/* note the cull table size command */
 		if (memcmp(cp, "culltable", 9) == 0 && isspace(cp[9])) {
 			unsigned long cts;
@@ -408,6 +451,44 @@ int main(int argc, char *argv[])
 			continue;
 		}
 
+		/* Note the suspension resume released file count thresholds
+		 * ("-" to disable a threshold).
+		 */
+		if (memcmp(cp, "resume_thresholds", 18) == 0 && isspace(cp[18])) {
+			unsigned long long b_thresh, f_thresh;
+			char *sp;
+
+			for (sp = cp + 18; isspace(*sp); sp++) {;}
+
+			if (*sp == '-') {
+				sp++;
+				b_thresh = ULLONG_MAX;
+			} else {
+				b_thresh = strtoul(sp, &sp, 10);
+			}
+
+			if (!*sp || !isspace(*sp))
+				cfgerror("Error parsing resume threshold (blocks)");
+			if (b_thresh == 0)
+				cfgerror("Invalid resume threshold (blocks)");
+			for (; isspace(*sp); sp++) {;}
+
+			if (*sp == '-') {
+				sp++;
+				f_thresh = ULLONG_MAX;
+			} else {
+				f_thresh = strtoul(sp, &sp, 10);
+				if (*sp)
+					cfgerror("Error parsing resume threshold (files)");
+				if (f_thresh == 0)
+					cfgerror("Invalid resume threshold (files)");
+			}
+
+			b_resume_threshold = b_thresh;
+			f_resume_threshold = f_thresh;
+			continue;
+		}
+
 		/* note the dir command */
 		if (memcmp(cp, "dir", 3) == 0 && isspace(cp[3])) {
 			char *sp;
@@ -448,13 +529,15 @@ int main(int argc, char *argv[])
 		oserror("Unable to close %s", configfile);
 
 	/* allocate the cull tables */
-	cullbuild = calloc(culltable_size, sizeof(cullbuild[0]));
-	if (!cullbuild)
-		oserror("calloc");
+	if (!culling_disabled) {
+		cullbuild = calloc(culltable_size, sizeof(cullbuild[0]));
+		if (!cullbuild)
+			oserror("calloc");
 
-	cullready = calloc(culltable_size, sizeof(cullready[0]));
-	if (!cullready)
-		oserror("calloc");
+		cullready = calloc(culltable_size, sizeof(cullready[0]));
+		if (!cullready)
+			oserror("calloc");
+	}
 
 	/* leave stdin, stdout, stderr and cachefd open only */
 	if (nullfd != 0)
@@ -468,7 +551,7 @@ int main(int argc, char *argv[])
 	/* set up a connection to syslog whilst we still can (the bind command
 	 * will give us our own namespace with no /dev/log */
 	openlog("cachefilesd", LOG_PID, LOG_DAEMON);
-	xopenedlog = 1;
+	xopenedlog = true;
 	info("About to bind cache");
 
 	/* now issue the bind command */
@@ -506,8 +589,7 @@ int main(int argc, char *argv[])
 	}
 
 	exit(0);
-
-} /* end main() */
+}
 
 /*****************************************************************************/
 /*
@@ -537,13 +619,12 @@ static void open_cache(void)
 	if (fstatfs(graveyardfd, &sfs) < 0)
 		oserror("Unable to stat cache filesystem");
 
-	if (sfs.f_bsize == -1 ||
-	    sfs.f_blocks == -1 ||
-	    sfs.f_bfree == -1 ||
-	    sfs.f_bavail == -1)
+	if (sfs.f_bsize  + 1 == 0 ||
+	    sfs.f_blocks + 1 == 0 ||
+	    sfs.f_bfree  + 1 == 0 ||
+	    sfs.f_bavail + 1 == 0)
 		error("Backing filesystem returns unusable statistics through fstatfs()");
-
-} /* end open_cache() */
+}
 
 /*****************************************************************************/
 /*
@@ -552,6 +633,8 @@ static void open_cache(void)
 static void cachefilesd(void)
 {
 	sigset_t sigs, osigs;
+	bool scanning_suspended = false;
+	bool scan_in_progress = false;
 
 	struct pollfd pollfds[1] = {
 		[0] = {
@@ -565,13 +648,14 @@ static void cachefilesd(void)
 	/* open the cache directories */
 	open_cache();
 
-	/* we need to disable I/O and termination signals so they're only
-	 * caught at appropriate times
+	/* We need to be able to disable signals that we need to check for
+	 * before calling poll so that we don't race and miss something.
 	 */
 	sigemptyset(&sigs);
 	sigaddset(&sigs, SIGIO);
 	sigaddset(&sigs, SIGINT);
 	sigaddset(&sigs, SIGTERM);
+	sigaddset(&sigs, SIGALRM);
 
 	signal(SIGTERM, sigterm);
 	signal(SIGINT, sigterm);
@@ -579,16 +663,81 @@ static void cachefilesd(void)
 	/* check the graveyard for graves */
 	reap_graveyard();
 
-	while (!stop) {
+	while (!stop_signalled) {
+		bool do_cull = false;
+
+		debug(3, "Loop %sbuild=%d ready=%d susp=%u scan=%u",
+		      culling_disabled ? "NOCULL " : "",
+		      nr_in_build_table, nr_in_ready_table,
+		      scanning_suspended, scan_in_progress);
+
 		read_cache_state();
 
-		/* sleep without racing on reap and cull with the signal
-		 * handlers */
-		if (!scan && !reap && !cull) {
+		if (!culling_disabled) {
+			/* Determine if we're going to need to start a new scan
+			 * to refill the cull table.  We want to do this if the
+			 * secondary cull table is less than half full - but
+			 * overriding that, we don't want to do this if we know
+			 * there's insufficient cullables to make it worth
+			 * while.
+			 */
+			if (!scan_in_progress) {
+				bool begin_scan = false;
+
+				debug(1, "Consider scan %d/%d",
+				      nr_in_build_table, culltable_size / 2);
+
+				if (nr_in_build_table < culltable_size / 2) {
+					debug(1, "Want to scan");
+					begin_scan = true;
+				}
+
+				if (begin_scan && scanning_suspended) {
+					debug(1, "Scanning suspended");
+					if (have_nr_releases) {
+						if (f_released_since_last_scan <
+						    f_resume_threshold &&
+						    b_released_since_last_scan <
+						    b_resume_threshold)
+							begin_scan = false;
+					} else {
+						begin_scan = scan_signalled;
+					}
+				}
+
+				if (begin_scan) {
+					debug(1, "Beginning a scan");
+					begin_building_cull_table();
+					scan_in_progress = true;
+					scanning_suspended = false;
+					scan_signalled = false;
+					f_released_since_last_scan = 0;
+					b_released_since_last_scan = 0;
+				}
+			}
+
+			/* Determine if there's anything we can actually cull yet if
+			 * the kernel is calling for space.
+			 */
+			if (kernel_wants_cull) {
+				debug(1, "Want to cull");
+				if (nr_in_ready_table > 0)
+					do_cull = true;
+			}
+		}
+
+		/* We block the signals across the checks for reap, cull and
+		 * scan initiation before polling so that we sleep without
+		 * racing against the signal handlers.
+		 */
+		if (!scan_in_progress && !reap_signalled && !do_cull) {
 			if (sigprocmask(SIG_BLOCK, &sigs, &osigs) < 0)
 				oserror("Unable to block signals");
 
-			if (!reap && !cull) {
+			if (!reap_signalled &&
+			    !stop_signalled &&
+			    !scan_signalled) {
+				debug(1, "Poll");
 				if (ppoll(pollfds, 1, NULL, &osigs) < 0 &&
 				    errno != EINTR)
 					oserror("Unable to suspend process");
@@ -596,40 +745,50 @@ static void cachefilesd(void)
 
 			if (sigprocmask(SIG_UNBLOCK, &sigs, NULL) < 0)
 				oserror("Unable to unblock signals");
-
-			read_cache_state();
+			continue;
 		}
 
-		if (jumpstart_scan) {
-			jumpstart_scan = 0;
-			if (!stop && !scan) {
-				notice("Refilling cull table");
-				root.usage++;
-				scan = &root;
+		if (!culling_disabled) {
+			if (do_cull)
+				cull_objects();
+
+			if (scan_in_progress) {
+				scan_in_progress = build_cull_table();
+				if (!scan_in_progress) {
+					/* Scan complete.
+					 *
+					 * If the scan didn't produce a full
+					 * table then don't repeat the scan
+					 * until something gets released by the
+					 * kernel.
+					 */
+					if (nr_in_build_table < culltable_size) {
+						debug(1, "Suspend scanning");
+						scanning_suspended = true;
+						if (!have_nr_releases) {
+							signal(SIGALRM, sigalrm);
+							alarm(30);
+						}
+					}
+				}
+			}
+
+			if (!scan_in_progress) {
+				if (nr_in_ready_table <= culltable_size / 2 + 2 &&
+				    nr_in_build_table > 0) {
+					debug(1, "Decant");
+					decant_cull_table();
+				}
 			}
 		}
 
-		if (cull) {
-			if (oldest_ready >= 0)
-				cull_objects();
-			else if (oldest_build < 0)
-				jumpstart_scan = 1;
-		}
-
-		if (scan)
-			build_cull_table();
-
-		if (!scan && oldest_ready < 0 && oldest_build >= 0)
-			decant_cull_table();
-
-		if (reap)
+		if (reap_signalled)
 			reap_graveyard();
 	}
 
 	notice("Daemon Terminated");
 	exit(0);
-
-} /* end cachefilesd() */
+}
 
 /*****************************************************************************/
 /*
@@ -638,14 +797,13 @@ static void cachefilesd(void)
 static void reap_graveyard(void)
 {
 	/* set a one-shot notification to catch more graves appearing */
-	reap = 0;
+	reap_signalled = false;
 	signal(SIGIO, sigio);
 	if (fcntl(graveyardfd, F_NOTIFY, DN_CREATE) < 0)
 		oserror("unable to set notification on graveyard");
 
 	reap_graveyard_aux(graveyardpath);
-
-} /* end reap_graveyard() */
+}
 
 /*****************************************************************************/
 /*
@@ -654,8 +812,9 @@ static void reap_graveyard(void)
 static void reap_graveyard_aux(const char *dirname)
 {
 	struct dirent dirent, *de;
+	bool deleted;
 	DIR *dir;
-	int deleted, ret;
+	int ret;
 
 	if (chdir(dirname) < 0)
 		oserror("chdir failed");
@@ -668,7 +827,7 @@ static void reap_graveyard_aux(const char *dirname)
 		/* removing directory entries may cause us to skip when reading
 		 * them */
 		rewinddir(dir);
-		deleted = 0;
+		deleted = false;
 
 		while (ret = readdir_r(dir, &dirent, &de),
 		       ret == 0 && de != NULL
@@ -682,7 +841,7 @@ static void reap_graveyard_aux(const char *dirname)
 					continue;
 			}
 
-			deleted = 1;
+			deleted = true;
 
 			/* attempt to unlink non-directory files */
 			if (dirent.d_type != DT_DIR) {
@@ -713,8 +872,7 @@ static void reap_graveyard_aux(const char *dirname)
 
 	if (chdir("..") < 0)
 		oserror("Unable to chdir to ..");
-
-} /* end reap_graveyard_aux() */
+}
 
 /*****************************************************************************/
 /*
@@ -730,6 +888,8 @@ static void read_cache_state(void)
 		oserror("Unable to read cache state");
 	buffer[n] = '\0';
 
+	debug(3, "KERNEL: %s", buffer);
+
 	tok = buffer;
 	do {
 		next = strpbrk(tok, " \t");
@@ -737,27 +897,37 @@ static void read_cache_state(void)
 			*next++ = '\0';
 
 		arg = strchr(tok, '=');
-		if (arg)
+		if (arg) {
 			*arg++ = '\0';
+		} else {
+			debug(0, "Warning: malformed output from kernel, missing arg to [%s]", tok);
+			continue;
+		}
 
-		if (strcmp(tok, "cull") == 0)
-			cull = strtoul(arg, NULL, 0);
-		else if (strcmp(tok, "brun") == 0)
+		if (strcmp(tok, "cull") == 0) {
+			kernel_wants_cull = (strtoul(arg, NULL, 0) != 0);
+		} else if (strcmp(tok, "brun") == 0) {
 			brun = strtoull(arg, NULL, 16);
-		else if (strcmp(tok, "bcull") == 0)
+		} else if (strcmp(tok, "bcull") == 0) {
 			bcull = strtoull(arg, NULL, 16);
-		else if (strcmp(tok, "bstop") == 0)
+		} else if (strcmp(tok, "bstop") == 0) {
 			bstop = strtoull(arg, NULL, 16);
-		else if (strcmp(tok, "frun") == 0)
+		} else if (strcmp(tok, "frun") == 0) {
 			frun = strtoull(arg, NULL, 16);
-		else if (strcmp(tok, "fcull") == 0)
+		} else if (strcmp(tok, "fcull") == 0) {
 			fcull = strtoull(arg, NULL, 16);
-		else if (strcmp(tok, "fstop") == 0)
+		} else if (strcmp(tok, "fstop") == 0) {
 			fstop = strtoull(arg, NULL, 16);
+		} else if (strcmp(tok, "breleased") == 0) {
+			b_released_since_last_scan += strtoull(arg, NULL, 16);
+			have_nr_releases = true;
+		} else if (strcmp(tok, "freleased") == 0) {
+			f_released_since_last_scan += strtoull(arg, NULL, 16);
+			have_nr_releases = true;
+		}
 
 	} while ((tok = next));
-
-} /* end read_cache_state() */
+}
 
 /*****************************************************************************/
 /*
@@ -776,8 +946,7 @@ static int is_object_in_use(const char *filename)
 		oserror("Failed to check object's in-use state");
 
 	return ret < 0 && errno == EBUSY ? 1 : 0;
-
-} /* end is_object_in_use */
+}
 
 /*****************************************************************************/
 /*
@@ -795,8 +964,7 @@ static void cull_file(const char *filename)
 	ret = write(cachefd, buffer, n);
 	if (ret < 0 && errno != ESTALE && errno != ENOENT && errno != EBUSY)
 		oserror("Failed to cull object");
-
-} /* end cull_file() */
+}
 
 /*****************************************************************************/
 /*
@@ -835,7 +1003,7 @@ static struct object *create_object(struct object *parent,
 		oserror("Unable to alloc object");
 
 	object->usage = 1;
-	object->new = 1;
+	object->new = true;
 
 	object->ino = st->st_ino;
 	object->atime = st->st_atime;
@@ -876,8 +1044,7 @@ static struct object *create_object(struct object *parent,
 
 	nobjects++;
 	return object;
-
-} /* end create_object() */
+}
 
 /*****************************************************************************/
 /*
@@ -922,8 +1089,7 @@ static void put_object(struct object *object)
 
 	if (parent)
 		put_object(parent);
-
-} /* end put_object() */
+}
 
 /*****************************************************************************/
 /*
@@ -953,8 +1119,7 @@ static void destroy_unexpected_object(struct object *parent, struct dirent *de)
 			oserror("Unable to rename unexpectedly named file: %s",
 				de->d_name);
 	}
-
-} /* end destroy_unexpected_object() */
+}
 
 /*****************************************************************************/
 /*
@@ -968,21 +1133,21 @@ static void insert_into_cull_table(struct object *object)
 		error("NULL object pointer");
 
 	/* just insert if table is empty */
-	if (oldest_build == -1) {
+	if (nr_in_build_table == 0) {
 		object->usage++;
-		oldest_build = 0;
 		cullbuild[0] = object;
+		nr_in_build_table++;
 		return;
 	}
 
 	/* insert somewhere if table is not full */
-	if (oldest_build < culltable_size - 1) {
+	if (nr_in_build_table < culltable_size) {
 		object->usage++;
-		oldest_build++;
 
 		/* just insert at end if new oldest object */
-		if (object->atime <= cullbuild[oldest_build - 1]->atime) {
-			cullbuild[oldest_build] = object;
+		if (object->atime <= cullbuild[nr_in_build_table - 1]->atime) {
+			cullbuild[nr_in_build_table] = object;
+			nr_in_build_table++;
 			return;
 		}
 
@@ -990,25 +1155,27 @@ static void insert_into_cull_table(struct object *object)
 		if (object->atime > cullbuild[0]->atime) {
 			memmove(&cullbuild[1],
 				&cullbuild[0],
-				oldest_build * sizeof(cullbuild[0]));
+				nr_in_build_table * sizeof(cullbuild[0]));
 
 			cullbuild[0] = object;
+			nr_in_build_table++;
 			return;
 		}
 
 		/* if only two objects in list then insert between them */
-		if (oldest_build == 2) {
+		if (nr_in_build_table == 2) {
 			cullbuild[2] = cullbuild[1];
 			cullbuild[1] = object;
+			nr_in_build_table++;
 			return;
 		}
 
 		/* insert somewhere in between front and back elements
-		 * of a three object list
+		 * of a three-plus object list
 		 * - oldest_build == #objects_currently_in_list
 		 */
 		y = 1;
-		o = oldest_build - 1;
+		o = nr_in_build_table - 1;
 
 		do {
 			m = (y + o) / 2;
@@ -1022,14 +1189,15 @@ static void insert_into_cull_table(struct object *object)
 
 		memmove(&cullbuild[y + 1],
 			&cullbuild[y],
-			(oldest_build - y) * sizeof(cullbuild[0]));
+			(nr_in_build_table - y) * sizeof(cullbuild[0]));
 
 		cullbuild[y] = object;
+		nr_in_build_table++;
 		return;
 	}
 
 	/* if table is full then insert only if older than newest */
-	if (oldest_build > culltable_size - 1)
+	if (nr_in_build_table > culltable_size)
 		error("Cull table overfull");
 
 	if (object->atime >= cullbuild[0]->atime)
@@ -1085,24 +1253,36 @@ static void insert_into_cull_table(struct object *object)
 		(y - 2) * sizeof(cullbuild[0]));
 
 	cullbuild[y - 1] = object;
-
-} /* end insert_into_cull_table() */
+}
 
 /*****************************************************************************/
 /*
- * do the next step in building up the cull table
+ * Begin a scan to build a cull table.
  */
-static void build_cull_table(void)
+static void begin_building_cull_table(void)
+{
+	debug(1, "Refilling cull table");
+	root.usage++;
+	scan_cursor = &root;
+}
+
+/*****************************************************************************/
+/*
+ * Do the next step in building up the cull table.  Returns false upon
+ * completion of a scan.
+ */
+static bool build_cull_table(void)
 {
 	struct dirent dirent, *de;
 	struct object *curr, *child;
 	struct stat64 st;
-	int loop, fd;
+	unsigned loop;
+	int fd;
 
-	curr = scan;
+	curr = scan_cursor;
 
 	if (!curr->dir) {
-		curr->empty = 1;
+		curr->empty = true;
 
 		fd = openat(dirfd(curr->parent->dir), curr->name, O_DIRECTORY);
 		if (fd < 0) {
@@ -1175,7 +1355,7 @@ next:
 	if (!child && errno == ENOENT)
 		goto next;
 
-	curr->empty = 0;
+	curr->empty = false;
 
 	if (!child)
 		oserror("Unable to create object");
@@ -1198,45 +1378,41 @@ next:
 				goto next;
 			}
 
-			for (loop = 0; loop <= oldest_ready; loop++)
+			for (loop = 0; loop < nr_in_ready_table; loop++)
 				if (cullready[loop] == child)
 					break;
 
-			if (loop == oldest_ready) {
+			if (loop == nr_in_ready_table - 1) {
 				/* child was oldest object */
-				cullready[oldest_ready] = (void *)(0x6b000000 | __LINE__);
-				oldest_ready--;
+				cullready[--nr_in_ready_table] = (void *)(0x6b000000 | __LINE__);
 				put_object(child);
 				goto removed;
 			}
-			else if (loop < oldest_ready) {
+			else if (loop < nr_in_ready_table - 1) {
 				/* child was somewhere in between */
 				memmove(&cullready[loop],
 					&cullready[loop + 1],
-					(oldest_ready - loop) * sizeof(cullready[0]));
-				cullready[oldest_ready] = (void *)(0x6b000000 | __LINE__);
-				oldest_ready--;
+					(nr_in_ready_table - (loop + 1)) * sizeof(cullready[0]));
+				cullready[--nr_in_ready_table] = (void *)(0x6b000000 | __LINE__);
 				put_object(child);
 				goto removed;
 			}
 
-			for (loop = 0; loop <= oldest_build; loop++)
+			for (loop = 0; loop < nr_in_build_table; loop++)
 				if (cullbuild[loop] == child)
 					break;
 
-			if (loop == oldest_build) {
+			if (loop == nr_in_build_table - 1) {
 				/* child was oldest object */
-				cullbuild[oldest_build] = (void *)(0x6b000000 | __LINE__);
-				oldest_build--;
+				cullbuild[--nr_in_build_table] = (void *)(0x6b000000 | __LINE__);
 				put_object(child);
 			}
-			else if (loop < oldest_build) {
+			else if (loop < nr_in_build_table - 1) {
 				/* child was somewhere in between */
 				memmove(&cullbuild[loop],
 					&cullbuild[loop + 1],
-					(oldest_build - loop) * sizeof(cullbuild[0]));
-				cullbuild[oldest_build] = (void *)(0x6b000000 | __LINE__);
-				oldest_build--;
+					(nr_in_build_table - (loop + 1)) * sizeof(cullbuild[0]));
+				cullbuild[--nr_in_build_table] = (void *)(0x6b000000 | __LINE__);
 				put_object(child);
 			}
 
@@ -1247,7 +1423,7 @@ next:
 		/* add objects that aren't in use to the cull table */
 		if (!is_object_in_use(dirent.d_name)) {
 			debug(2, "- insert");
-			child->new = 0;
+			child->new = false;
 			insert_into_cull_table(child);
 		}
 		put_object(child);
@@ -1258,11 +1434,11 @@ next:
 	case OBJTYPE_INTERMEDIATE:
 		debug(2, "- descend");
 
-		child->new = 0;
-		scan = child;
+		child->new = false;
+		scan_cursor = child;
 
 		debug(2, "<-- build_cull_table({%s})", curr->name);
-		return;
+		return true;
 
 	default:
 		error("Unexpected type");
@@ -1305,15 +1481,13 @@ dir_read_complete:
 		}
 	}
 
-	scan = curr->parent;
-	if (!scan) {
-		info("Scan complete");
-		decant_cull_table();
-	}
+	scan_cursor = curr->parent;
+	if (!scan_cursor)
+		debug(1, "Scan complete");
 
 	debug(2, "<-- build_cull_table({%s})", curr->name);
 	put_object(curr);
-	return;
+	return scan_cursor != NULL;
 
 	/* delete unexpected objects that we've found */
 found_unexpected_object:
@@ -1321,8 +1495,7 @@ found_unexpected_object:
 
 	destroy_unexpected_object(curr, &dirent);
 	goto next;
-
-} /* end build_cull_table() */
+}
 
 /*****************************************************************************/
 /*
@@ -1330,76 +1503,66 @@ found_unexpected_object:
  */
 static void decant_cull_table(void)
 {
-	int loop, space, avail, copy, leave, n;
+	unsigned loop, avail, copy, leave, space, n;
 
-	if (scan)
+	if (scan_cursor)
 		error("Can't decant cull table whilst scanning");
 
-	/* if nothing there, scan again in a short while */
-	if (oldest_build < 0) {
-		signal(SIGALRM, sigalrm);
-		alarm(30);
-		return;
-	}
-
 	/* mark the new entries cullable */
-	for (loop = 0; loop <= oldest_build; loop++) {
+	for (loop = 0; loop < nr_in_build_table; loop++) {
 		if (!cullbuild[loop]->cullable) {
-			cullbuild[loop]->cullable = 1;
+			cullbuild[loop]->cullable = true;
 			ncullable++;
 		}
 	}
 
 	/* if the ready table is empty, copy the whole lot across */
-	if (oldest_ready == -1) {
-		copy = oldest_build + 1;
+	if (nr_in_ready_table == 0) {
+		copy = nr_in_build_table;
 
-		info("Decant (all %d)", copy);
+		debug(1, "Decant (all %d)", copy);
 
 		n = copy * sizeof(cullready[0]);
 		memcpy(cullready, cullbuild, n);
 		memset(cullbuild, 0x6e, n);
-		oldest_ready = oldest_build;
-		oldest_build = -1;
+		nr_in_ready_table = nr_in_build_table;
+		nr_in_build_table = 0;
 		goto check;
 	}
 
 	/* decant some of the build table if there's space */
-	space = culltable_size - (oldest_ready + 1);
-	if (space <= 0) {
-		if (space < 0)
-			error("Less than zero space in ready table");
+	if (culltable_size < nr_in_ready_table)
+		error("Less than zero space in ready table");
+	space = culltable_size - nr_in_ready_table;
+	if (space == 0)
 		goto check;
-	}
 
 	/* work out how much of the build table we can copy */
-	copy = avail = oldest_build + 1;
+	copy = avail = nr_in_build_table;
 	if (copy > space)
 		copy = space;
 	leave = avail - copy;
 
-	info("Decant (%d/%d to %d)", copy, avail, space);
+	debug(1, "Decant (%u/%u to %u)", copy, avail, space);
 
 	/* make a hole in the ready table transfer "copy" elements from the end
 	 * of cullbuild (oldest) to the beginning of cullready (youngest)
 	 */
-	n = oldest_ready + 1;
-	memmove(&cullready[copy], &cullready[0], n * sizeof(cullready[0]));
-	oldest_ready += copy;
+	memmove(&cullready[copy], &cullready[0], nr_in_ready_table * sizeof(cullready[0]));
+	nr_in_ready_table += copy;
 
 	memcpy(&cullready[0], &cullbuild[leave], copy * sizeof(cullready[0]));
 	memset(&cullbuild[leave], 0x6b, copy * sizeof(cullbuild[0]));
-	oldest_build = leave - 1;
+	nr_in_build_table = leave;
 
 	if (copy + leave > culltable_size)
 		error("Scan table exceeded (%d+%d)", copy, leave);
 
 check:
-	for (loop = 0; loop < oldest_ready; loop++)
+	for (loop = 0; loop < nr_in_ready_table; loop++)
 		if (((long)cullready[loop] & 0xf0000000) == 0x60000000)
 			abort();
-
-} /* end decant_cull_table() */
+}
 
 /*****************************************************************************/
 /*
@@ -1429,8 +1592,7 @@ static int get_dir_fd(struct object *dir)
 	debug(1, "<%d>/%s to %d", parentfd, dir->name, fd);
 	close(parentfd);
 	return fd;
-
-} /* end get_dir_fd() */
+}
 
 /*****************************************************************************/
 /*
@@ -1463,8 +1625,7 @@ static void cull_object(struct object *object)
 
 object_already_gone:
 	put_object(object);
-
-} /* end cull_object() */
+}
 
 /*****************************************************************************/
 /*
@@ -1475,19 +1636,8 @@ static void cull_objects(void)
 	if (ncullable <= 0)
 		error("Cullable object count is inconsistent");
 
-	if (cullready[oldest_ready]->cullable) {
-		cull_object(cullready[oldest_ready]);
-		cullready[oldest_ready] = (void *)(0x6b000000 | __LINE__);
-		oldest_ready--;
+	if (cullready[nr_in_ready_table - 1]->cullable) {
+		cull_object(cullready[nr_in_ready_table - 1]);
+		cullready[--nr_in_ready_table] = (void *)(0x6b000000 | __LINE__);
 	}
-
-	/* must start refilling the cull table */
-	if (!scan && oldest_build <= culltable_size / 2 + 2) {
-		decant_cull_table();
-
-		notice("Refilling cull table");
-		root.usage++;
-		scan = &root;
-	}
-
-} /* end cull_objects() */
+}
